#include "FS2_Flag.h"
#include "IniParser.h"
#include "Settings.h"
#include "System.h"
#include "FilePaths.h"

createFileSingleton( Settings );


bool Settings::initialize()
{
	/* lets create config dir and data dir inside it if they dont exist */
	if( FILEPATHS.getGameDir() != FILEPATHS.getConfigDir() )
	{
		if( ! QFile::exists( FILEPATHS.getConfigDir() ) )
			System::mkDir( FILEPATHS.getConfigDir() );

		if( ! QFile::exists( FILEPATHS.getConfigDataDir() ) )
			System::mkDir( FILEPATHS.getConfigDataDir() );
	}

	IniParser ipLauncher6;

	/* if there is no laucher6.ini we try global one, then we bail */
	if( ! ipLauncher6.loadIni( FILEPATHS.getConfigLauncherIni() ) )
	{
		if( ! ipLauncher6.loadIni( FILEPATHS.getGameLauncherIni() ) )
		{
			qWarning( "launcher6.ini not found, select exec using browse button." );
			return false;
		}
	}

	QString qstrExeFilePath = ipLauncher6.getEntry( "launcher:exe_filepath" );
	return setExeFilePath( qstrExeFilePath );
}

bool Settings::setExeFilePath( const QString & qstrExeFilePath )
{
	QString qstrExe = QFileInfo( qstrExeFilePath ).canonicalFilePath();

	/* check if the exe is in our game dir */
	if( QFileInfo( qstrExe ).exists() && QFileInfo( qstrExe ).path() != FILEPATHS.getGameDir() )
	{
		System::warning( "The launcher must be in the same directory as the binary you are trying to use. Trying to find binary in the launcher directory instead." );

		/* try to find it in our current dir */
		qstrExe = FILEPATHS.getGameDir() + '/' + QFileInfo( qstrExeFilePath ).fileName();
	}

	/*
	 * the following should never happen as we select file by file
	 * dialog, the only option is to modify launcher6.ini by hand
	 * or fall into the above case with trying to find exe
	 */
	if( ! QFileInfo( qstrExe ).exists() )
	{
		qWarning( "Exe does not exist, select proper one using browse button." );
		return false;
	}
	if( ! QFileInfo( qstrExe ).isExecutable() )
	{
		qWarning( "Exe is not executable, select proper one using browse button." );
		return false;
	}

	/* new and seems to be correct qstrExe, everything that was before is superfluous */
	m_qstrExeFilePath = QDir::toNativeSeparators( qstrExe );
	m_qlFlags.clear();
	m_qstrlFlagTypes.clear();
	m_qstrlEasyFlagSets.clear();
	m_qstrCustomFlags = "";

	/* mod info should be cleared as well */
	m_qstrModName = "";
	m_qstrModFullName = "";
	m_qstrlModsPrimary.clear();
	m_qstrlModsSecondary.clear();
	m_qstrModImage = m_qstrModInfo = m_qstrModWebsite = m_qstrModForum = "";

	if( ! parseFlagList() )
		return false;

	IniParser ipLauncher6;

	/* this can fail, we don't need this file now */
	if( ! ipLauncher6.loadIni( FILEPATHS.getConfigLauncherIni() ) )
		ipLauncher6.loadIni( FILEPATHS.getGameLauncherIni() );

	QString qstrFlags = ipLauncher6.getEntry( "launcher:game_flags" );
	/*
	 * either game_flags is empty or whole launcher6.ini
	 * does not exist, lets try to read cmdline_fso.cfg
	 * to fill initial flag states, we need to ignore
	 * "-mod" and one param after cause of this, we
	 * can't use mod value to fill setModName() in
	 * such case cause there might be more then one mod
	 */
	if( qstrFlags.isEmpty() )
	{
		QFile qfCmdLine( FILEPATHS.getConfigCmdLine() );
		if( qfCmdLine.open( QIODevice::ReadOnly | QIODevice::Text ) )
		{
			qstrFlags = qfCmdLine.readLine().simplified();
			qfCmdLine.close();
		}
	}
	QStringList qstrlFlags = qstrFlags.split( " " );

	for( int i = 0; i < qstrlFlags.size(); i++ )
	{
		if( qstrlFlags.at( i ) == "-mod" )
		{
			i++;
			continue;
		}
		/*
		 * we don't know whether it's custom flag or not
		 * setFlag() must support custom flags
		 */
		setFlag( qstrlFlags.at( i ), true, true );
	}

	QString qstrModName = ipLauncher6.getEntry( "launcher:active_mod" );
	return setModName( qstrModName );
}

bool Settings::parseFlagList()
{
	int nEasyFlagSize;
	int nFlagSize;

	int nNumEasyFlags;
	EasyFlag *pEasyFlags = NULL;
	int nNumFlags;
	Flag *pFlags = NULL;

	QFile qfFlags( FILEPATHS.getFlagsLch() );
	qfFlags.remove();

	/* let's go to tmp dir, current dir is not guaranteed for rw */
	QString qstrCurrentDir = System::chDir( FILEPATHS.getTempDir() );
	if( qstrCurrentDir.isEmpty() )
	{
		qWarning( "CRITICAL: cannot change dir to tmp." );
		return false;
	}
	/* run the game in tmp dir */
	if( ! System::run( m_qstrExeFilePath, QStringList( "-get_flags" ), true ) )
	{
		qWarning( "CRITICAL: cannot exec game to test flags." );
		return false;
	}
	/* go back to previous dir */
	System::chDir( qstrCurrentDir );

	qfFlags.open( QIODevice::ReadOnly );
	if ( ! qfFlags.isOpen() )
	{
		qWarning( "CRITICAL: Cannot open flags.lch file." );
		goto bail;
	}

	qfFlags.read( ( char* )( &nEasyFlagSize ), sizeof( int ) );
	qfFlags.read( ( char* )( &nFlagSize ), sizeof( int ) );
	if( nEasyFlagSize != sizeof( EasyFlag ) || nFlagSize != sizeof( Flag ) )
	{
		qWarning( "CRITICAL: fs2_open and launcher versions don't match." );
		goto bail;
	}

	qfFlags.read( ( char* )( &nNumEasyFlags ), sizeof( int ) );
	pEasyFlags = new EasyFlag[ nNumEasyFlags ];
	if( pEasyFlags == NULL )
	{
		qWarning( "CRITICAL: Cannot allocate memory." );
		goto bail;
	}
	qfFlags.read( ( char* )pEasyFlags, sizeof( EasyFlag ) * nNumEasyFlags );

	qfFlags.read( ( char* )( &nNumFlags ), sizeof( int ) );
	pFlags = new Flag[ nNumFlags ];
	if( pFlags == NULL )
	{
		qWarning( "CRITICAL: Cannot allocate memory." );
		goto bail;
	}
	qfFlags.read( ( char* )pFlags, sizeof( Flag ) * nNumFlags );

	/* lets rewrite those ugly :-) C structs to something more friendly */
	for( int i = 0; i < nNumFlags; i++ )
	{
		CFlag cfNewFlag;
		cfNewFlag.m_qstrFlag = pFlags[ i ].name;
		cfNewFlag.m_qstrDesc = pFlags[ i ].desc;
		cfNewFlag.m_nEasyFlagsOn = pFlags[ i ].on_flags;
		cfNewFlag.m_nEasyFlagsOff = pFlags[ i ].off_flags;
		cfNewFlag.m_qstrType = pFlags[ i ].type;
		cfNewFlag.m_qstrUrl = pFlags[ i ].web_url;
		cfNewFlag.m_bState = false;

		if( m_qstrlFlagTypes.indexOf( cfNewFlag.m_qstrType ) == -1 )
			m_qstrlFlagTypes.append( cfNewFlag.m_qstrType );

		m_qlFlags.append( cfNewFlag );
	}

	for( int i = 0; i < nNumEasyFlags; i++ )
		m_qstrlEasyFlagSets.append( pEasyFlags[ i ].name );

	//qWarning( "easy num: %d num: %d", nNumEasyFlags, nNumFlags );
	//qWarning( "easy size: %d size: %d", nEasyFlagSize, nFlagSize );

	delete pEasyFlags;
	delete pFlags;
	qfFlags.close();
	qfFlags.remove();
	return true;

bail:
	if( pEasyFlags != NULL ) delete pEasyFlags;
	if( pFlags != NULL ) delete pFlags;
	qfFlags.close();
	qfFlags.remove();
	return false;
}

bool Settings::setModName( const QString & qstrModName )
{
	/*
	 * the following should never happen as we select mod by custom
	 * dialog, the only option is to modify launcher6.ini by hand
	 */
	if( ! QFile::exists( FILEPATHS.getModDir( qstrModName ) ) )
	{
		qWarning( "Mod dir does not exist. Using base game." );
		m_qstrModName = "";
	}
	else
		m_qstrModName = qstrModName;

	m_qstrModFullName = "";
	m_qstrlModsPrimary.clear();
	m_qstrlModsSecondary.clear();
	m_qstrModImage = m_qstrModInfo = m_qstrModWebsite = m_qstrModForum = "";

	IniParser ipTmp;
	QString qstrTmp;
	QStringList qstrlTmp;

#if 0
	/* let's try to load settings.ini, ignore it's custom flags */
	if( ipTmp.loadIni( FILEPATHS.getModSettingsIni( m_qstrModName ) ) )
	{
		qstrTmp = ipTmp.getEntry( "settings:flags" );
		qstrlTmp = qstrTmp.split( " ", QString::SkipEmptyParts );
		for( int i = 0; i < qstrlTmp.size(); i++ )
			setFlag( qstrlTmp.at( i ), true );
	}
#endif

	/*
	 * this mod has no ini file, nothing more to do here
	 * (it doesn't mean that it's not correct)
	 */
	if( ! ipTmp.loadIni( FILEPATHS.getModIni( m_qstrModName ) ) )
		return false;

	m_qstrModFullName = ipTmp.getEntry( "launcher:modname" );
	m_qstrModImage = ipTmp.getEntry( "launcher:image255x112" );
	m_qstrModInfo = ipTmp.getEntry( "launcher:infotext" );
	m_qstrModWebsite = ipTmp.getEntry( "launcher:website" );
	m_qstrModForum = ipTmp.getEntry( "launcher:forum" );

	/* let's try to read flags from mod.ini, ignore custom flags */
	qstrTmp = ipTmp.getEntry( "settings:flags" );
	qstrlTmp = qstrTmp.split( " ", QString::SkipEmptyParts );
	for( int i = 0; i < qstrlTmp.size(); i++ )
		setFlag( qstrlTmp.at( i ), true );

	qstrTmp = ipTmp.getEntry( "multimod:primarylist" );
	qstrlTmp = qstrTmp.split( ",", QString::SkipEmptyParts );
	for( int i = 0; i < qstrlTmp.size(); i++ )
		m_qstrlModsPrimary.append( qstrlTmp.at( i ) );

	qstrTmp = ipTmp.getEntry( "multimod:secondrylist" );
	qstrlTmp = qstrTmp.split( ",", QString::SkipEmptyParts );
	qstrTmp = ipTmp.getEntry( "multimod:secondarylist" );
	qstrlTmp += qstrTmp.split( ",", QString::SkipEmptyParts );
	for( int i = 0; i < qstrlTmp.size(); i++ )
		m_qstrlModsSecondary.append( qstrlTmp.at( i ) );

	return true;
}

void Settings::setFlag( const QString & qstrFlag, bool bState, bool bSetCustom )
{
	int i;

	for( i = 0; i < m_qlFlags.size(); i++ )
		if( m_qlFlags.at( i ).m_qstrFlag == qstrFlag )
			break;

	if( i < m_qlFlags.size() )
		m_qlFlags[ i ].m_bState = bState;
	/*
	 * someone tries to set single custom flag using this function
	 * let's append it at the end without any checks
	 */
	else if( bState && bSetCustom )
	{
		m_qstrCustomFlags.append( " " );
		m_qstrCustomFlags.append( qstrFlag );
		m_qstrCustomFlags = m_qstrCustomFlags.simplified();
	}
}

void Settings::setCustomFlags( const QString & qstrCustomFlags )
{
	m_qstrCustomFlags = qstrCustomFlags.simplified();
}

QString Settings::generateCmdLine( bool bWithExe, bool bWithMod ) const
{
	QString qstrArgs;

	if( bWithExe )
		qstrArgs = m_qstrExeFilePath;

	if( bWithMod )
		if( ! m_qstrModName.isEmpty() )
		{
			qstrArgs.append( " -mod " );
			for( int i = 0 ; i < m_qstrlModsPrimary.size(); i++ )
			{
				qstrArgs.append( m_qstrlModsPrimary.at( i ) );
				qstrArgs.append( "," );
			}
			qstrArgs.append( m_qstrModName );
			for( int i = 0 ; i < m_qstrlModsSecondary.size(); i++ )
			{
				qstrArgs.append( "," );
				qstrArgs.append( m_qstrlModsSecondary.at( i ) );
			}
			qstrArgs.append( " " );
		}

	for( int i = 0; i < m_qlFlags.size(); i++ )
		if( m_qlFlags.at( i ).m_bState == true )
		{
			qstrArgs.append( " " );
			qstrArgs.append( m_qlFlags.at( i ).m_qstrFlag );
		}

	qstrArgs.append( " " );
	qstrArgs.append( m_qstrCustomFlags );

	return qstrArgs.simplified();
}

void Settings::setEasyTypeSet( int nType )
{
	if( nType == 0 )
		return;

	int nFlag = 0;
	switch( nType )
	{
		case 1:
			nFlag = EASY_DEFAULT;
			break;
		case 2:
			nFlag = EASY_ALL_ON;
			break;
		case 3:
			nFlag = EASY_MEM_ON;
			break;
		case 4:
			nFlag = EASY_MEM_OFF;
			break;
	}

	for( int i = 0; i < m_qlFlags.size(); i++ )
	{
		if( ( m_qlFlags.at( i ).m_nEasyFlagsOn & nFlag ) == nFlag )
			m_qlFlags[ i ].m_bState = true;

		if( ( m_qlFlags.at( i ).m_nEasyFlagsOff & nFlag ) == nFlag )
			m_qlFlags[ i ].m_bState = false;
	}
}

bool Settings::save() const
{
	IniParser ipLauncher6;
	ipLauncher6.setEntry( "launcher:exe_filepath", m_qstrExeFilePath );
	ipLauncher6.setEntry( "launcher:game_flags", generateCmdLine( false, false ) );
	ipLauncher6.setEntry( "launcher:active_mod", m_qstrModName );
	if( ! ipLauncher6.dumpIni( FILEPATHS.getConfigLauncherIni() ) )
	{
		qWarning( "CRITICAL: saving launcher6.ini failed." );
		return false;
	}

	QFile qfCmdLine( FILEPATHS.getConfigCmdLine() );
	if( ! qfCmdLine.open( QIODevice::WriteOnly | QIODevice::Text ) )
	{
		qWarning( "CRITICAL: saving cmdline_fso.cfg line failed." );
		return false;
	}
	QTextStream qtsCmdLine( &qfCmdLine );
	qtsCmdLine << generateCmdLine( false, true ) << endl;
	qfCmdLine.close();

	return true;
}

const QString & Settings::getExeFilePath() const
{
	return m_qstrExeFilePath;
}

const QString & Settings::getCustomFlags() const
{
	return m_qstrCustomFlags;
}

const QFlagList & Settings::getFlags() const
{
	return m_qlFlags;
}

const QStringList & Settings::getFlagTypes() const
{
	return m_qstrlFlagTypes;
}

const QStringList & Settings::getEasyFlagSets() const
{
	return m_qstrlEasyFlagSets;
}

const QString & Settings::getModName() const
{
	return m_qstrModName;
}

const QString & Settings::getModFullName() const
{
	if( ! m_qstrModFullName.isEmpty() )
		return m_qstrModFullName;

	return getModName();
}

const QString & Settings::getModImage() const
{
	return m_qstrModImage;
}

const QString & Settings::getModInfo() const
{
	return m_qstrModInfo;
}

const QString & Settings::getModWebsite() const
{
	return m_qstrModWebsite;
}

const QString & Settings::getModForum() const
{
	return m_qstrModForum;
}
